#include "string_matcher.h"
#include <sstream>

using namespace std;

namespace string_matcher {

namespace {

struct Quoted
{
    const string& value;
};

ostream& operator<<(ostream& os, Quoted s)
{
    os << '"';
    for (char c : s.value) {
        if (c == '"' || c == '\\')
            os << '\\';
        os << c;
    }
    os << '"';
    return os;
}

class Conjunction_matcher : public IMatcher_impl
{
public:
    Conjunction_matcher(const String_matcher&, const String_matcher&);
    bool matches(const string&) const override;
    void describe(ostream&) const override;
private:
    String_matcher left_, right_;
};

Conjunction_matcher::Conjunction_matcher(
        const String_matcher& left, const String_matcher& right)
        : left_(left), right_(right)
{ }

bool Conjunction_matcher::matches(const string& actual) const
{
    return left_.matches(actual) && right_.matches(actual);
}

void Conjunction_matcher::describe(ostream& os) const {
    os << "(and " << left_ << ' ' << right_ << ')';
}

class Disjunction_matcher : public IMatcher_impl
{
public:
    Disjunction_matcher(const String_matcher&, const String_matcher&);
    bool matches(const string&) const override;
    void describe(ostream&) const override;
private:
    String_matcher left_, right_;
};

Disjunction_matcher::Disjunction_matcher(
        const String_matcher& left, const String_matcher& right)
        : left_(left), right_(right)
{ }

bool Disjunction_matcher::matches(const string& actual) const
{
    return left_.matches(actual) || right_.matches(actual);
}

void Disjunction_matcher::describe(ostream& os) const {
    os << "(or " << left_ << ' ' << right_ << ')';
}

class Negation_matcher : public IMatcher_impl
{
public:
    Negation_matcher(const String_matcher&);
    bool matches(const string&) const override;
    void describe(ostream&) const override;
private:
    String_matcher negation_;
};

Negation_matcher::Negation_matcher(const String_matcher& negation)
        : negation_(negation)
{ }

bool Negation_matcher::matches(const string& actual) const
{
    return !negation_.matches(actual);
}

void Negation_matcher::describe(ostream& os) const {
    os << "(not " << negation_ << ')';
}

class Equals_matcher : public IMatcher_impl
{
public:
    explicit Equals_matcher(const string&);
    bool matches(const string&) const override;
    void describe(ostream&) const override;
private:
    string expected_;
};

string_matcher::Equals_matcher::Equals_matcher(const string& expected)
        : expected_(expected)
{ }

bool string_matcher::Equals_matcher::matches(const string& actual) const
{
    return actual == expected_;
}

void string_matcher::Equals_matcher::describe(ostream& os) const
{
    os << "(equals " << Quoted{expected_} << ")";
}

class Substring_at_matcher : public IMatcher_impl
{
public:
    Substring_at_matcher(const string&, size_t);
    bool matches(const string&) const override;
    void describe(ostream&) const override;

protected:
    string substring_;
    size_t position_;
};

bool Substring_at_matcher::matches(const string& actual) const
{
    if (position_ + substring_.size() > actual.size())
        return false;

    for (size_t i = 0; i < substring_.size(); ++i) {
        if (substring_[i] != actual[position_ + i])
            return false;
    }

    return true;
}

Substring_at_matcher::Substring_at_matcher(
        const string& substring, size_t position)
        : substring_(substring), position_(position)
{ }

void Substring_at_matcher::describe(ostream& os) const
{
    os << "(substring-at " << Quoted{substring_} << ' ' << position_ << ')';
}

class Starts_with_matcher : public Substring_at_matcher
{
public:
    explicit Starts_with_matcher(const string&);
    void describe(ostream& os) const override;
};

Starts_with_matcher::Starts_with_matcher(const string& prefix)
        : Substring_at_matcher(prefix, 0)
{}

void Starts_with_matcher::describe(ostream& os) const
{
    os << "(starts-with " << Quoted{substring_} << ')';
}

class Ends_with_matcher : public IMatcher_impl
{
public:
    explicit Ends_with_matcher(const string&);
    bool matches(const string&) const override;
    void describe(ostream&) const override;

private:
    string suffix_;
};

Ends_with_matcher::Ends_with_matcher(const string& suffix)
        : suffix_(suffix)
{ }

bool Ends_with_matcher::matches(const string& actual) const
{
    if (actual.size() < suffix_.size())
        return false;

    size_t offset = actual.size() - suffix_.size();

    return Substring_at_matcher(suffix_, offset).matches(actual);
}

void Ends_with_matcher::describe(ostream& os) const
{
    os << "(ends-with " << Quoted{suffix_} << ')';
}

class Contains_matcher : public IMatcher_impl
{
public:
    explicit Contains_matcher(const string&);
    bool matches(const string&) const override;
    void describe(ostream&) const override;

private:
    string substring_;
};

Contains_matcher::Contains_matcher(const string& substring)
        : substring_(substring)
{ }

bool Contains_matcher::matches(const string& actual) const
{
    return search(actual.begin(), actual.end(),
                  substring_.begin(), substring_.end())
           != actual.end();
}

void Contains_matcher::describe(ostream& os) const
{
    os << "(contains " << Quoted{substring_} << ')';
}

} // end namespace

String_matcher String_matcher::operator&&(const String_matcher& other) const
{
    return String_matcher::make<Conjunction_matcher>(*this, other);
}

String_matcher String_matcher::operator||(const String_matcher& other) const
{
    return String_matcher::make<Disjunction_matcher>(*this, other);
}

String_matcher String_matcher::operator!() const
{
    return String_matcher::make<Negation_matcher>(*this);
}

string String_matcher::to_string() const
{
    ostringstream ss;
    ss << *this;
    return ss.str();
}

String_matcher equals(const string& expected)
{
    return String_matcher::make<Equals_matcher>(expected);
}

String_matcher substring_at(const string& substring, size_t position)
{
    return String_matcher::make<Substring_at_matcher>(substring, position);
}

String_matcher starts_with(const string& prefix)
{
    return String_matcher::make<Starts_with_matcher>(prefix);
}

String_matcher ends_with(const string& suffix)
{
    return String_matcher::make<Ends_with_matcher>(suffix);
}

String_matcher contains(const string& substring)
{
    return String_matcher::make<Contains_matcher>(substring);
}

}